import javax.swing.*;
import java.awt.*;
import jfd.exp.*;

public class DrawPanel extends JPanel
{
   private static final double RADIANS_PER_DEGREE = Math.PI / 180.0;
   private static final double FACTOR1            = .4;
   private static final double FACTOR2            = .05;
   private static final double FACTOR3            = 1.1;

   private double axisLength = 0;

   private double cosX       = 0.0;
   private double cosY       = 0.0;
   private double cosZ       = 0.0;
   private double sinX       = 0.0;
   private double sinY       = 0.0;
   private double sinZ       = 0.0;
   private int    xOffset    = 0;
   private int    yOffset    = 0;

   private Point values[][]  = null;

   public DrawPanel()
   {
      setPreferredSize(new Dimension(0, 400));

      setBorder(
         BorderFactory.createCompoundBorder(
            BorderFactory.createEmptyBorder(2, 2, 2, 2),
            BorderFactory.createEtchedBorder()
         )
      );
   }

   public void draw(
      int xDegrees,
      int yDegrees,
      int zDegrees,
      int nDivisionsX,
      int nDivisionsY,
      double xStart,
      double yStart,
      double xEnd,
      double yEnd,
      Node n,
      DrawErrorObserver observer
   )
   {
      // Set values for 3D => 2D transformation

      cosX        = Math.cos(xDegrees * RADIANS_PER_DEGREE);
      cosY        = Math.cos(yDegrees * RADIANS_PER_DEGREE);
      cosZ        = Math.cos(zDegrees * RADIANS_PER_DEGREE);
      sinX        = Math.sin(xDegrees * RADIANS_PER_DEGREE);
      sinY        = Math.sin(yDegrees * RADIANS_PER_DEGREE);
      sinZ        = Math.sin(zDegrees * RADIANS_PER_DEGREE);

      // Create our values array

      double xStep = (xEnd - xStart) / (double)nDivisionsX;
      double yStep = (yEnd - yStart) / (double)nDivisionsY;

      ++nDivisionsX;
      ++nDivisionsY;

      values = new Point[nDivisionsX][nDivisionsY];

      SymbolTable table = new SymbolTable();

      try
      {
         int i = 0;

         for(double x = xStart; x <= xEnd; x += xStep, i ++)
         {
            table.setValue("x", x);

            int j = 0;

            for(double y = yStart; y <= yEnd; y += yStep, j ++)
            {
               table.setValue("y", y);

               double z = n.execute(table);

               values[i][j] = to2D(x, y, z);
            }
         }
      }
      catch(Exception e)
      {
         observer.drawErrorOccurred(e.getMessage());
         values = null;
         return;
      }

      repaint();
   }

   private Point to2D(double x, double y, double z)
   {
      double vx = x * cosX + y * cosY + z * cosZ;
      double vy = x * sinX + y * sinY + z * sinZ;

      return(new Point((int)(vx + 0.5), (int)(vy + 0.5)));
   }

   private Point xlate(Point p)
   {
      Point newPoint = new Point(p);
      newPoint.translate(xOffset, yOffset);
      return(newPoint);
   }

   public void paint(Graphics g)
   {
      super.paint(g);

      if(values == null)
         return;

      Dimension d = getSize();
      xOffset     = d.width  / 2;
      yOffset     = d.height / 2;
      axisLength  = FACTOR1 * Math.min(d.width, d.height);

      g.setColor(Color.blue);

      drawAxes(g);

      g.setColor(Color.red);

      Point p1 = null;
      Point p2 = null;

      for(int i = 0; i < values.length; i ++)
      {
         for(int j = 1; j < values[i].length; j ++)
         {
           p1 = xlate(values[i][j - 1]);
           p2 = xlate(values[i][j]);

            g.drawLine(p1.x, p1.y, p2.x, p2.y);
         }
      }

      for(int j = 0; j < values[0].length; j ++)
      {
         for(int i = 1; i < values.length; i ++)
         {
            p1 = xlate(values[i - 1][j]);
            p2 = xlate(values[i][j]);

            g.drawLine(p1.x, p1.y, p2.x, p2.y);
         }
      }
   }

   private void drawAxes(Graphics g)
   {
      Point center = xlate(to2D(0.0,        0.0,        0.0));
      Point xEnd   = xlate(to2D(axisLength, 0.0,        0.0));
      Point yEnd   = xlate(to2D(0.0,        axisLength, 0.0));
      Point zEnd   = xlate(to2D(0.0,        0.0,        axisLength));

      drawAxis(center, xEnd, "x", g);
      drawAxis(center, yEnd, "y", g);
      drawAxis(center, zEnd, "z", g);
   }

   private void drawAxis(Point center, Point end, String name, Graphics g)
   {
      g.drawLine(center.x, center.y, end.x, end.y);

      // First draw an arrowhead

      double deltaX = center.x - end.x;
      double deltaY = center.y - end.y;

      double theta = 0.0;

      if(deltaX == 0.0)
         theta = (deltaY >= 0.0 ? Math.PI / 2.0 : 3.0 * Math.PI / 2.0);

      else if(deltaY == 0.0)
         theta = (deltaX >= 0.0 ? 0.0 : Math.PI);
      
      else
         theta = Math.atan2(deltaY, deltaX);

      double r = Math.sqrt(deltaX * deltaX + deltaY * deltaY);

      double distance = r * FACTOR2;
      double angle    = theta + 20 * RADIANS_PER_DEGREE;

      Point side1 = new Point(
         (int)(end.x + distance * Math.cos(angle) + 0.5),
         (int)(end.y + distance * Math.sin(angle) + 0.5)
      );

      angle = theta - 20 * RADIANS_PER_DEGREE;

      Point side2 = new Point(
         (int)(end.x + distance * Math.cos(angle) + 0.5),
         (int)(end.y + distance * Math.sin(angle) + 0.5)
      );

      g.drawLine(end.x, end.y, side1.x, side1.y);
      g.drawLine(end.x, end.y, side2.x, side2.y);

      // Now label the axis

      theta   += Math.PI;
      distance = r * FACTOR3;

      Point p = new Point(
         (int)(center.x + distance * Math.cos(theta) + 0.5),
         (int)(center.y + distance * Math.sin(theta) + 0.5)
      );

      g.drawString(name, p.x, p.y);
   }
}
